//---------------------------------
// 2022/11/7

#include "InfraredRec.h"



//---------------------------------
// Private Class Function


// 赤外線受光ユニットのON -> OFFの入力を監視して、それぞれの長さを返す
//  引数
//   timeout : タイムアウト[us] この時間ONが来なければreturn
//  戻り値
//   Low 16bit : ON pulse [us] / 10
//   High 16bit : OFF pulse [us] / 10
//   0xFFFF : タイムアウト
uint32_t InfraredRec::receivePulse(uint32_t timeout){
  uint16_t to=0;
  uint16_t on=0;
  uint16_t off=0;

  // Wait ON pulse
  while(1==digitalRead(recGPIO)){
    to++;
    if(to>timeout){
      return 0x0000FFFF;
    }
    delayMicroseconds(1);
  }

  // ON pulse
  while(0==digitalRead(recGPIO)){
    delayMicroseconds(9);
    on++;
  }
  uint16_t l = on*8;  // Timeout limit
  
  // OFF pulse
  while(1==digitalRead(recGPIO)){
    delayMicroseconds(9);
    off++;
    if(off>l){
      break;
    }
  }

  uint32_t r = off;
  r = r <<16;
  r = r | on;
  return r;
}


// 赤外線フォーマットの判定 receivePulseの結果を引数にして呼び出す
uint8_t InfraredRec::checkFormat(uint32_t firstPulse){
  uint16_t on = firstPulse & 0xFFFF;
  uint16_t off = firstPulse >> 16;
  
  if(on>280 && on<450){
    return formatAEHA;
  }
  if(on>750 && on<1200){
    return formatNEC;
  }
  if(on>180 && on<280){
    return formatSONY;
  }
  return formatUnknown;
}

// リピートコードかどうかの判定 receivePulseの結果を引数にして呼び出す
bool InfraredRec::checkRepeat(uint8_t format, uint32_t firstPulse){
  uint16_t off = firstPulse >> 16;
  
  switch(format){
  case formatAEHA:
    if(off>270){
      return true;
    }
    return false;
  case formatNEC:
    if(off<338){
      return true;
    }
    return false;
  }
  return false;
}



// 1フレーム分のデータを受信してバッファbufに入れる NEC and AEHA format のみ
//  引数
//   timeout : 最初のONが来るまでのタイムアウト [us]
//  戻り値
//   リードしたデータのバイト数
//   0xFE : フォーマット判定に失敗
//   0xFF : タイムアウトもしくはリピートコード
uint8_t InfraredRec::readFrame(uint8_t*buf, uint32_t timeout){
  uint16_t length0;
  uint16_t length1;
  uint16_t lengthStop;
  uint8_t data=0;
  uint8_t byte=0;
  uint8_t b=0;
  
  uint32_t p = receivePulse(timeout);
  if(p==0xFFFF){
    return readTimeout;
  }
  
  format = checkFormat(p);
  if(checkRepeat(format, p)){
    return(readUnknown);
  }
  
  switch(format){
  case formatAEHA:
    length0=42;
    length1=127;
    lengthStop=200;
    break;
  case formatNEC:
    length0=56;
    length1=169;
    lengthStop=260;
    break;
  default:
    return readUnknown;
  }
  
  while(1){
    p = receivePulse(50000);
    // Timeout
    if(p==0xFFFF){
      break;
    }
    
    p = p>>16;  // OFF length
    if((length0*0.5) < p && (length0*1.5) > p){
      ;
    }else if((length1*0.5) < p && (length1*1.5) > p){
      data = data | (1<<b);  // LSB first
    }else if(lengthStop < p){
      break;
    }
    b++;
    
    // Final bit of 1 byte
    if(b==8){
      buf[byte] = data;
      data=0;
      b=0;
      byte++;
      
      if(byte>=maxFrameSize){
        return byte;
      }
    }
  }
  return byte;
}


//---------------------------------
// Public Class Function

// 入力端子の番号で初期化
InfraredRec::InfraredRec(uint8_t setGPIO){
  recGPIO = setGPIO;
  pinMode(recGPIO, INPUT);

  format = formatUnknown;
  frameCount = 0;
}


// 赤外線データ受信 (複数フレーム対応)
//  戻り値
//   0 : 成功
//   0xFE : フォーマット判定エラー
//   0xFF : タイムアウト
uint8_t InfraredRec::readData(){
  for(uint8_t c = 0; c<maxFrameCount; c++){
    if(c==0){
      frameSize[c] = readFrame(frameData[c], 10);
      
      // Timeout or non supported format
      if(frameSize[c] == readTimeout || frameSize[c] == readUnknown){
        return frameSize[c];
      }
    }else{
      frameSize[c] = readFrame(frameData[c], 50000);
      
      // Timeout
      if(frameSize[c] == readTimeout || frameSize[c] == readUnknown){
        return readSuccess;
      }
    }

    frameCount = c+1;
  }
  return readSuccess;
}

